import re
from ..utils import int_to_roman, get_best_match
# import fuzzy
# import phonetics

class WikipediaBaseFilter(object):
    def description_has_keywords(self, description):
        if any(x in description for x in self.keywords):
            return True
        return False

    def description_has_name(self, description, name):
        # If name is in the description, then return true
        if name in description:
            return True
        # Otherwise, find the best matching string.
        match, value = get_best_match(name, description)
        # If it matches 0.94 or higher, try to check the characters.
        if value >= 0.94:
            # Find the characters that are different.
            diff = tuple(set(name).symmetric_difference(set(match)))
            # Save whetehre each different character is alphanumeric or not
            # ie: A-Z or ':' or '_'
            alphanumeric = []
            for char in diff:
                if char.isalpha() or char.isnumeric():
                    alphanumeric.append((char, True))
                else:
                    alphanumeric.append((char, False))
            # If it only contains extra non-alphanumeric characters,
            # then consider it a match, and return true.
            if all(t[1] is False for t in alphanumeric):
                return True
        # Otherwise, return false.
        return False

    def url_has_name(self, url, name):
        name_lower = name.lower()
        name_lower_underscore = name_lower.replace(" ", "_")
        name_lower_slug = name_lower.replace(" ", "-")
        return any(x in url.lower() for x in [
            name_lower,
            name_lower_underscore,
            name_lower_slug,
        ])

    def clean_description(self, description):
        description = re.sub(r"[^\x00-\x7F]", " ", description)
        description = re.sub(r"\n", " ", description)
        description = re.sub("\'", "", description)
        description = re.sub("'", "", description)
        description = re.sub(r"\s\s+", " ", description)
        return description

    def filter(self, film_name, scraped_data):
        result = None
        while len(scraped_data) != 0:
            item = scraped_data.pop(0)
            if result is None:
                url = item["url"]
                raw_description = item["description"]
                if raw_description is None:
                    continue
                cleaned_description = self.clean_description(raw_description)
                if self.description_has_name(cleaned_description, film_name):
                    if self.description_has_keywords(cleaned_description):
                        result = item
                else:
                    try:
                        m = re.search(r"\d+", film_name)
                        number = m.group(0)
                        roman = int_to_roman(int(number))
                        roman_name = film_name.replace(number, roman)
                        roman_name_lower = film_name.replace(number, roman.lower())
                        if self.decription_has_name(cleaned_description, roman_name):
                            if self.description_has_keywords(cleaned_description):
                                result = item
                        if self.decription_has_name(cleaned_description, roman_name_lower):
                            if self.description_has_keywords(cleaned_description):
                                result = item
                        if self.url_has_name(url, roman_name):
                            result = item
                        if self.url_has_name(url, roman_name_lower):
                            result = item
                    except AttributeError:
                        pass
                if self.url_has_name(url, film_name):
                    result = item
        return result




import json
import re
import requests
import bs4
from .filters.film import WikipediaFilmFilter


class WikipediaScraper(object):
    def __init__(self, string,
                       limit=10,
                       disambiguation_id="Film_and_television",
                       disambiguation_keyword="film",
                       filter_class=WikipediaFilmFilter):
        # the string to scrape for.
        self.string = string
        # number of items to return from scraping.
        self.limit = limit
        # Id of the disambiguation element to get anchors from.
        self.disambiguation_id = disambiguation_id
        # The keyword to search for in the disambiguated links list.
        self.disambiguation_keyword = disambiguation_keyword
        # The raw data we scraped from wikipedia
        self.scraped = None
        # The filter class to use on the scraped data
        self.filter_class = filter_class
        self.IMAGE_FILETYPES = (
            'jpg', 'jpeg', 'gif', 'png',
            "JPG", "JPEG", "GIF", "PNG",
        )
        self.SKIPPED_FILETYPES = tuple([
            "svg.%s" % filetype
            for filetype in self.IMAGE_FILETYPES
        ])
        self.RE_WIKIPEDIA_ANNOTATIONS = r'(\[[0-9]+\])'
        self.IGNORED_WIKIPEDIA_DESCRIPTIONS = (
            'this article contains a list of works that does not follow the manual of style for lists of works',
        )
        self.WIKIMEDIA_UPLOAD_URL = "//upload.wikimedia.org"
        self.WIKIPEDIA_URL = "https://en.wikipedia.org"
        self.WIKIPEDIA_API_URL = "w/api.php?action=opensearch&search={}&limit={}&namespace=0&format=json"

    def get_soup(self, url):
        req = requests.get(url)
        if req.status_code != 200:
            raise Exception("Failed to download page.")
        soup = bs4.BeautifulSoup(req.content, 'html.parser')
        return soup

    def filter_images(self, soup):
        images = []
        for tag in soup.findAll("img"):
            src = tag["src"]
            if src.endswith(self.IMAGE_FILETYPES):
                if not src.endswith(self.SKIPPED_FILETYPES):
                    if src.startswith(self.WIKIMEDIA_UPLOAD_URL):
                        images.append(src)
        return images

    def filter_description(self, soup):
        try:
            infobox = soup.findAll('table')[0]
            desc = infobox.findNextSibling()
            if desc is not None:
                desc = desc.getText().strip().lower()
                desc = re.sub(self.RE_WIKIPEDIA_ANNOTATIONS, '', desc)
                if not any(x in desc for x in self.IGNORED_WIKIPEDIA_DESCRIPTIONS):
                    return desc
        except IndexError:
            return ""

    def filter_disambiguated(self, soup):
        urls = []
        try:
            anchors = soup.find(id=self.disambigation_id).find_next("ul").find_all("a")
            for anchor in anchors:
                for k, v in anchor.attrs.items():
                    if k == "href":
                        if self.disambiguation_keyword in v:
                            if not v.startswith(self.WIKIPEDIA_URL):
                                v = "{}{}".format(self.WIKIPEDIA_URL, v)
                            urls.append(v)
        except AttributeError:
            pass
        return urls

    def is_disambiguated(self, soup):
        title = soup.find("title")
        results = re.search(r"\(disambiguation\)", title.text)
        try:
            result = results.group()
            return True
        except AttributeError:
            # Didn't match anything.
            return False

    def make_data(self, url, soup):
        return {
            "url": url,
            "description": self.filter_description(soup),
            "images": self.filter_images(soup),
        }

    def scrape(self, limit=None):
        if limit is None:
            limit = self.limit
        scraped_pages = []
        for url in self.make_urls(limit):
            for scraped_page in self.scrape_page(url):
                scraped_pages.append(scraped_page)
        self.scraped = scraped_pages
        return scraped_pages

    def filter(self):
        filter_instance = self.filter_class()
        filtered = filter_instance.filter(self.string, self.scraped)
        return filtered

    def scrape_page(self, url):
        scraped = []
        original_soup = self.get_soup(url)
        if self.is_disambiguated(original_soup):
            # add if condition for different disambiguation paths
            # ie for: films -> if string ends with 'film'
            #         music -> if string ends with 'music'
            for ft_page_url in self.filter_disambiguated(original_soup):
                ft_page_soup = self.get_soup(ft_page_url)
                ft_page_data = self.make_data(ft_page_url, ft_page_soup)
                scraped.append(ft_page_data)
        else:
            scraped.append(self.make_data(url, original_soup))
        return scraped

    def make_urls(self, limit):
        # Make a url for for a page to scrape.
        query_string = self.string.strip()
        query_string = query_string.replace(r"\s\s+", " ")
        query_string = query_string.replace(" ", "%20")
        query_url = "{}/{}".format(
            self.WIKIPEDIA_URL,
            self.WIKIPEDIA_API_URL.format(
                query_string,
                limit,
            )
        )
        req = requests.get(query_url)
        text = req.text
        # Convert the data into a list of urls
        data = json.loads(text)
        all_urls = data[-1]
        # split the search term into parts
        bits = query_string.lower().split()
        suffix = bits.pop()
        prefix = ' '.join(bits)
        prefix_underscore = prefix.replace(" ", "_")
        prefix_slug = prefix.replace(" ", "-")
        # Create a list of urls that have the film name in it.
        prefixes = [prefix, prefix_underscore, prefix_slug]
        confirmed_urls = []
        for url in all_urls:
            if any(x in url.lower() for x in prefixes):
                confirmed_urls.append(url)
        return confirmed_urls



from .base import WikipediaBaseFilter


class WikipediaFilmFilter(WikipediaBaseFilter):
    keywords = [
        "film",
        "films",
        "movie",
        "movies",
        "theatrical release",
        "theatrical releases"
        "home video",
        "home videos",
    ]



import sys
from difflib import SequenceMatcher


def get_best_match(query,
                   corpus,
                   step=4,
                   flex=3,
                   case_sensitive=False,
                   verbose=False):
    """Return best matching substring of corpus.

    Parameters
    ----------
    query : str
    corpus : str
    step : int
        Step size of first match-value scan through corpus. Can be thought of
        as a sort of "scan resolution". Should not exceed length of query.
    flex : int
        Max. left/right substring position adjustment value. Should not
        exceed length of query / 2.

    Outputs
    -------
    output0 : str
        Best matching substring.
    output1 : float
        Match ratio of best matching substring. 1 is perfect match.

    https://stackoverflow.com/questions/36013295/find-best-substring-match
    """

    def ratio(a, b):
        """Compact alias for SequenceMatcher."""
        return SequenceMatcher(None, a, b).ratio()

    def scan_corpus(step):
        """Return list of match values from corpus-wide scan."""
        match_values = []

        m = 0
        while m + qlen - step <= len(corpus):
            match_values.append(ratio(query, corpus[m : m-1+qlen]))
            if verbose:
                print(query, "-", corpus[m: m + qlen], ratio(query, corpus[m: m + qlen]))
            m += step

        return match_values

    def index_max(v):
        """Return index of max value."""
        return max(range(len(v)), key=v.__getitem__)

    def adjust_left_right_positions():
        """Return left/right positions for best string match."""
        # bp_* is synonym for 'Best Position Left/Right' and are adjusted
        # to optimize bmv_*
        p_l, bp_l = [pos] * 2
        p_r, bp_r = [pos + qlen] * 2

        # bmv_* are declared here in case they are untouched in optimization
        bmv_l = match_values[int(p_l / step)]
        bmv_r = match_values[int(p_r / step)]

        for f in range(flex):
            ll = ratio(query, corpus[p_l - f: p_r])
            if ll > bmv_l:
                bmv_l = ll
                bp_l = p_l - f

            lr = ratio(query, corpus[p_l + f: p_r])
            if lr > bmv_l:
                bmv_l = lr
                bp_l = p_l + f

            rl = ratio(query, corpus[p_l: p_r - f])
            if rl > bmv_r:
                bmv_r = rl
                bp_r = p_r - f

            rr = ratio(query, corpus[p_l: p_r + f])
            if rr > bmv_r:
                bmv_r = rr
                bp_r = p_r + f

            if verbose:
                print("\n" + str(f))
                print("ll: -- value: {} -- snippet: {}".format(
                    ll, corpus[p_l - f: p_r]
                ))
                print("lr: -- value: {} -- snippet: {}".format(
                    lr, corpus[p_l + f: p_r]
                ))
                print("rl: -- value: {} -- snippet: {}".format(
                    rl, corpus[p_l: p_r - f]
                ))
                print("rr: -- value: {} -- snippet: {}".format(
                    rl, corpus[p_l: p_r + f]
                ))

        return bp_l, bp_r, ratio(query, corpus[bp_l : bp_r])

    if not case_sensitive:
        query = query.lower()
        corpus = corpus.lower()

    qlen = len(query)

    if flex >= qlen/2:
        print("Warning: flex exceeds length of query / 2. Setting to default.")
        flex = 3

    match_values = scan_corpus(step)
    pos = index_max(match_values) * step

    pos_left, pos_right, match_value = adjust_left_right_positions()

    return corpus[pos_left: pos_right].strip(), match_value


def int_to_roman(num):
    """
    https://stackoverflow.com/questions/42875103/integer-to-roman-number
    https://stackoverflow.com/questions/33486183/convert-from-numbers-to-roman-notation
    """
    conv = (
        ("M", 1000),
        ("CM", 900),
        ("D", 500),
        ("CD", 400),
        ("C", 100),
        ("XC", 90),
        ("L", 50),
        ("XL", 40),
        ("X", 10),
        ("IX", 9),
        ("V", 5),
        ("IV", 4),
        ("I", 1)
    )
    roman = ""
    i = 0
    while num > 0:
        while conv[i][1] > num:
            i += 1
        roman += conv[i][0]
        num -= conv[i][1]
    return roman


def roman_to_int(roman):
    """
    https://gist.github.com/kristopherjohnson/f4eca9018c5085fc736b2f29a202b8f8
    """
    vals = {
        "I": 1,
        "V": 5,
        "X": 10,
        "L": 50,
        "C": 100,
        "D": 500,
        "M": 1000,
    }
    total = 0
    lastValue = sys.maxsize
    for char in list(roman):
        value = vals[char.upper()]
        if value > lastValue:
            total += value - 2 * lastValue
        else:
            total += value
        lastValue = value
    return total



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# https://setuptools.readthedocs.io/en/latest/setuptools.html

from __future__ import absolute_import
from __future__ import print_function
from glob import glob
from setuptools import setup, find_packages
from os.path import abspath, basename, dirname, join, splitext
import re

HERE = abspath(dirname(__file__))
NAME = basename(HERE)


def read(*parts):
    with open(join(HERE, *parts), "r", encoding="utf-8") as fp:
        return fp.read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


setup(
    # This is the name of your project. The first time you publish this
    # package, this name will be registered for you. It will determine how
    # users can install this project, e.g.:
    #
    # $ pip install sampleproject
    #
    # And where it will live on PyPI: https://pypi.org/project/sampleproject/
    #
    # There are some restrictions on what makes a valid project name
    # specification here:
    # https://packaging.python.org/specifications/core-metadata/#name
    #
    # Required
    name=NAME,

    # Versions should comply with PEP 440:
    # https://www.python.org/dev/peps/pep-0440/
    #
    # For a discussion on single-sourcing the version across setup.py and the
    # project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    #
    # Required
    version=find_version("src", NAME, "__init__.py"),

    # This is a one-line description or tagline of what your project does. This
    # corresponds to the "Summary" metadata field:
    # https://packaging.python.org/specifications/core-metadata/#summary
    #
    # Required
    description="Scrapes pages like wikipedia for for urls, descriptions, and images",

    # This is an optional longer description of your project that represents
    # the body of text which users will see when they visit PyPI.
    #
    # Often, this is the same as your README, so you can just read it in from
    # that file directly (as we have already done above)
    #
    # This field corresponds to the "Description" metadata field:
    # https://packaging.python.org/specifications/core-metadata/#description-optional
    #
    # Optional
    long_description=read("README.md"),

    # Denotes that our long_description is in Markdown; valid values are
    # text/plain, text/x-rst, and text/markdown
    #
    # Optional if long_description is written in reStructuredText (rst) but
    # required for plain-text or Markdown; if unspecified, "applications should
    # attempt to render [the long_description] as text/x-rst; charset=UTF-8 and
    # fall back to text/plain if it is not valid rst" (see link below)
    #
    # This field corresponds to the "Description-Content-Type" metadata field:
    # https://packaging.python.org/specifications/core-metadata/#description-content-type-optional
    #
    # Optional (see note above)
    long_description_content_type="text/markdown",

    # This should be a valid link to your project"s main homepage.
    #
    # This field corresponds to the "Home-Page" metadata field:
    # https://packaging.python.org/specifications/core-metadata/#home-page-optional
    #
    # Optional
    url="https://github.com/alexseitsinger/{}".format(NAME),

    # This should be your name or the name of the organization which owns the
    # project.
    #
    # Optional
    author="Alex Seitsinger",

    # This should be a valid email address corresponding to the author listed
    # above.
    #
    # Optional
    author_email="alexseitsinger@gmail.com",

    # Classifiers help users find your project by categorizing it.
    #
    # classifier list: http://pypi.python.org/pypi?%3Aaction=list_classifiers
    #
    # Optional
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        "Development Status :: 3 - Alpha",

        # Indicate who your project is intended for
        "Intended Audience :: Developers",
        "Topic :: Utilities",

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        "Programming Language :: Python",
        # "Programming Language :: Python :: 2",
        # "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",

        # Pick your license as you wish
        "License :: OSI Approved :: BSD License",
    ],

    # This field adds keywords for your project which will appear on the
    # project page. What does your project relate to?
    #
    # Optional
    keywords=[
        # eg:
        #   "sample", "test"
    ],

    py_modules=[
        splitext(basename(path))[0] for path in glob("src/*.py")
    ],

    # You can just specify package directories manually here if your project is
    # simple. Or you can use find_packages().
    #
    # Alternatively, if you just want to distribute a single Python file, use
    # the `py_modules` argument instead as follows, which will expect a file
    # called `my_module.py` to exist:
    #
    #   py_modules=["my_module"],
    #
    # Required
    packages=find_packages("src"),

    # Tell distutils packages are under "./src"
    # Required if we store packages under a "src" directory.
    #
    # Optional
    package_dir={"": "src"},

    # A file or a string to specify the type of license for this package.
    #
    # Optional
    license="BSD 2-Clause License",

    # Whether the project can be safely installed and run from a zip file. If
    # this argument is not supplied, the bdist_egg command will have to
    # analyze all of your project’s contents for possible problems each time
    # it builds an egg.
    #
    # Optional
    zip_safe=False,

    # This field lists other packages that your project depends on to run.
    # Any package you put here will be installed by pip when your project is
    # installed, so they must be valid existing projects.
    #
    # For an analysis of "install_requires" vs pip"s requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    #
    # Optional
    install_requires=[
        # eg:
        #   "six", or "six>=1.7",
        "requests",
        "bs4",
    ],

    # List additional groups of dependencies here (e.g. development
    # dependencies). Users will be able to install these using the "extras"
    # syntax, for example:
    #
    #   $ pip install sampleproject[dev]
    #
    # Similar to `install_requires` above, these must be valid existing
    # projects.
    #
    # Optional
    extras_require={
        # eg:
        #   "rst": ["docutils>=0.11"],
        #   ":python_version=="2.6"": ["argparse"],
    },

    # If there are data files included in your packages that need to be
    # installed, specify them here.
    #
    # If using Python 2.6 or earlier, then these have to be included in
    # MANIFEST.in as well.
    #
    # You do not need to use this option if you are using
    # include_package_data, unless you need to add e.g. files that are
    # generated by your setup script and build process. (And are therefore not
    # in source control or are files that you don’t want to include in your
    # source distribution.)
    #
    # Optional
    package_data={
        # eg:
        #   "sample": ["package_data.dat"],
    },

    # This tells setuptools to automatically include any data files it finds
    # inside your package directories that are specified by your MANIFEST.in
    # file.
    include_package_data=True,

    # Although "package_data" is the preferred approach, in some case you may
    # need to place data files outside of your packages. See:
    # http://docs.python.org/3.4/distutils/setupscript.html#installing-additional-files
    #
    # In this case, "data_file" will be installed into "<sys.prefix>/my_data"
    #
    # Optional
    data_files=[
        # eg:
        #   ("my_data", ["data/data_file"]),
    ],

    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # `pip` to create the appropriate form of executable for the target
    # platform.
    #
    # For example, the following would provide a command called `sample` which
    # executes the function `main` from this package when invoked:
    #
    # Optional
    entry_points={
        # eg:
        #   "console_scripts": [
        #       "sample=sample:main",
        #   ],
    },

    # List additional URLs that are relevant to your project as a dict.
    #
    # This field corresponds to the "Project-URL" metadata fields:
    # https://packaging.python.org/specifications/core-metadata/#project-url-multiple-use
    #
    # Examples listed include a pattern for specifying where the package tracks
    # issues, where the source is hosted, where to say thanks to the package
    # maintainers, and where to support the project financially. The key is
    # what"s used to render the link text on PyPI.
    #
    # Optional
    project_urls={
        # eg:
        #   "Bug Reports": "https://github.com/pypa/sampleproject/issues",
        #   "Funding": "https://donate.pypi.org",
        #   "Say Thanks!": "http://saythanks.io/to/example",
        #   "Source": "https://github.com/pypa/sampleproject/",
    },
)
